@@ -12,6 +12,26 @@ pub const MAX_PER_SEC_RANDOM: u32 = 0xFF_FFFF;
 /// Digit characters used in the base 32 notation.
 const DIGITS: &[u8; 32] = b"0123456789ABCDEFGHIJKLMNOPQRSTUV";
 
+/// O(1) map from ASCII values to base 32 digit values.
+const DECODE_MAP: [u8; 256] = [
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,
+    0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,
+    0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+];
+
 /// Represents a SCRU128 ID and provides various converters and comparison operators.
 ///
 /// # Examples
@@ -97,15 +117,22 @@ impl FromStr for Scru128Id {
 
     /// Creates an object from a 26-digit string representation.
     fn from_str(str_value: &str) -> Result<Self, Self::Err> {
-        let mut cs = str_value.chars();
-        if str_value.len() == 26
-            && cs.next().map_or(false, |c| c.is_digit(8))
-            && cs.all(|c| c.is_digit(32))
-        {
-            Ok(Self(u128::from_str_radix(str_value, 32).unwrap()))
-        } else {
-            Err(ParseError(str_value.into()))
+        if str_value.len() != 26 {
+            return Err(ParseError {});
         }
+        let bs = str_value.as_bytes();
+        let mut int_value = DECODE_MAP[bs[0] as usize] as u128;
+        if int_value > 7 {
+            return Err(ParseError {});
+        }
+        for i in 1..26 {
+            let n = DECODE_MAP[bs[i] as usize] as u128;
+            if n == 0xff {
+                return Err(ParseError {});
+            }
+            int_value = (int_value << 5) | n;
+        }
+        Ok(Self(int_value))
     }
 }
 
@@ -150,11 +177,11 @@ impl From<Scru128Id> for String {
 
 /// Error parsing an invalid string representation of SCRU128 ID.
 #[derive(Clone, Eq, PartialEq, Hash, Debug)]
-pub struct ParseError(String);
+pub struct ParseError {}
 
 impl fmt::Display for ParseError {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "invalid string representation: {}", self.0)
+        write!(f, "invalid string representation")
     }
 }
 
@@ -165,35 +192,30 @@ mod tests {
     use super::Scru128Id;
     use crate::Scru128Generator;
 
+    const MAX_UINT44: u64 = (1 << 44) - 1;
+    const MAX_UINT28: u32 = (1 << 28) - 1;
+    const MAX_UINT24: u32 = (1 << 24) - 1;
+    const MAX_UINT32: u32 = u32::MAX;
+
     /// Encodes and decodes prepared cases correctly
     #[test]
     fn it_encodes_and_decodes_prepared_cases_correctly() {
         let cases: Vec<((u64, u32, u32, u32), &str)> = vec![
             ((0, 0, 0, 0), "00000000000000000000000000"),
-            ((2u64.pow(44) - 1, 0, 0, 0), "7VVVVVVVVG0000000000000000"),
-            ((2u64.pow(44) - 1, 0, 0, 0), "7vvvvvvvvg0000000000000000"),
-            ((0, 2u32.pow(28) - 1, 0, 0), "000000000FVVVVU00000000000"),
-            ((0, 2u32.pow(28) - 1, 0, 0), "000000000fvvvvu00000000000"),
-            ((0, 0, 2u32.pow(24) - 1, 0), "000000000000001VVVVS000000"),
-            ((0, 0, 2u32.pow(24) - 1, 0), "000000000000001vvvvs000000"),
-            ((0, 0, 0, u32::MAX), "00000000000000000003VVVVVV"),
-            ((0, 0, 0, u32::MAX), "00000000000000000003vvvvvv"),
+            ((MAX_UINT44, 0, 0, 0), "7VVVVVVVVG0000000000000000"),
+            ((MAX_UINT44, 0, 0, 0), "7vvvvvvvvg0000000000000000"),
+            ((0, MAX_UINT28, 0, 0), "000000000FVVVVU00000000000"),
+            ((0, MAX_UINT28, 0, 0), "000000000fvvvvu00000000000"),
+            ((0, 0, MAX_UINT24, 0), "000000000000001VVVVS000000"),
+            ((0, 0, MAX_UINT24, 0), "000000000000001vvvvs000000"),
+            ((0, 0, 0, MAX_UINT32), "00000000000000000003VVVVVV"),
+            ((0, 0, 0, MAX_UINT32), "00000000000000000003vvvvvv"),
             (
-                (
-                    2u64.pow(44) - 1,
-                    2u32.pow(28) - 1,
-                    2u32.pow(24) - 1,
-                    u32::MAX,
-                ),
+                (MAX_UINT44, MAX_UINT28, MAX_UINT24, MAX_UINT32),
                 "7VVVVVVVVVVVVVVVVVVVVVVVVV",
             ),
             (
-                (
-                    2u64.pow(44) - 1,
-                    2u32.pow(28) - 1,
-                    2u32.pow(24) - 1,
-                    u32::MAX,
-                ),
+                (MAX_UINT44, MAX_UINT28, MAX_UINT24, MAX_UINT32),
                 "7vvvvvvvvvvvvvvvvvvvvvvvvv",
             ),
         ];
@@ -241,7 +263,7 @@ mod tests {
     /// Returns error if an invalid string representation is supplied
     #[test]
     fn it_returns_error_if_an_invalid_string_representation_is_supplied() {
-        let cases = [
+        let cases = vec![
             "",
             " 00SCT4FL89GQPRHN44C4LFM0OV",
             "00SCT4FL89GQPRJN44C7SQO381 ",
@@ -263,22 +285,34 @@ mod tests {
         }
     }
 
-    /// Has symmetric converters from/to String, u128, and fields
+    /// Has symmetric converters from/to various values
     #[test]
     fn it_has_symmetric_converters() {
+        let mut cases = vec![
+            Scru128Id::from_fields(0, 0, 0, 0),
+            Scru128Id::from_fields(MAX_UINT44, 0, 0, 0),
+            Scru128Id::from_fields(0, MAX_UINT28, 0, 0),
+            Scru128Id::from_fields(0, 0, MAX_UINT24, 0),
+            Scru128Id::from_fields(0, 0, 0, MAX_UINT32),
+            Scru128Id::from_fields(MAX_UINT44, MAX_UINT28, MAX_UINT24, MAX_UINT32),
+        ];
+
         let mut g = Scru128Generator::new();
         for _ in 0..1000 {
-            let obj = g.generate();
-            assert_eq!(obj.to_string().parse::<Scru128Id>(), Ok(obj));
-            assert_eq!(Scru128Id::from_u128(obj.as_u128()), obj);
+            cases.push(g.generate());
+        }
+
+        for e in cases {
+            assert_eq!(e.to_string().parse::<Scru128Id>(), Ok(e));
+            assert_eq!(Scru128Id::from_u128(e.as_u128()), e);
             assert_eq!(
                 Scru128Id::from_fields(
-                    obj.timestamp(),
-                    obj.counter(),
-                    obj.per_sec_random(),
-                    obj.per_gen_random()
+                    e.timestamp(),
+                    e.counter(),
+                    e.per_sec_random(),
+                    e.per_gen_random()
                 ),
-                obj
+                e
             );
         }
     }
@@ -296,11 +330,11 @@ mod tests {
         let mut ordered = vec![
             Scru128Id::from_fields(0, 0, 0, 0),
             Scru128Id::from_fields(0, 0, 0, 1),
-            Scru128Id::from_fields(0, 0, 0, 0xFFFF_FFFF),
+            Scru128Id::from_fields(0, 0, 0, MAX_UINT32),
             Scru128Id::from_fields(0, 0, 1, 0),
-            Scru128Id::from_fields(0, 0, 0xFF_FFFF, 0),
+            Scru128Id::from_fields(0, 0, MAX_UINT24, 0),
             Scru128Id::from_fields(0, 1, 0, 0),
-            Scru128Id::from_fields(0, 0xFFF_FFFF, 0, 0),
+            Scru128Id::from_fields(0, MAX_UINT28, 0, 0),
             Scru128Id::from_fields(1, 0, 0, 0),
             Scru128Id::from_fields(2, 0, 0, 0),
         ];
